# AUTOGENERATED! DO NOT EDIT! File to edit: gaussian.ipynb (unless otherwise specified).

__all__ = ['sample_conditional', 'sample_bayes_net']

# Internal Cell
import gtsam
import numpy as np
import plotly.express as px
import pandas as pd

# Cell
def sample_conditional(node: gtsam.GaussianConditional, N: int, parents: list = [], sample: dict = {}):
    """Sample from conditional """
    # every node ~ exp(0.5*|R x + S p - d|^2)
    # calculate mean as inv(R)*(d - S p)
    d = node.d()
    n = len(d)
    rhs = d.reshape(n, 1)
    if len(parents) > 0:
        rhs = rhs - node.S() @ np.vstack([sample[p] for p in parents])
    # sample from conditional Gaussian
    invR = np.linalg.inv(node.R())
    return invR @ (rhs + np.random.normal(size=(n, N)))


def sample_bayes_net(bn: gtsam.GaussianBayesNet, N: int) -> dict:
    """ High performance ancestral sampling.
        It returns a dictionary of nj x N samples, where n_j is the dimensionality for key j.
    """
    sample = {}
    for i in reversed(range(bn.size())):
        conditional = bn.at(i)
        key, *parents = conditional.keys()
        sample[key] = sample_conditional(bn.at(i), N, parents, sample)
    return sample


# AUTOGENERATED! DO NOT EDIT! File to edit: html.ipynb (unless otherwise specified).

__all__ = ['td', 'tr', 'th', 'table', 'row', 'ROW']

# Cell
from IPython.display import HTML

# Cell
def td(obj):
    """Create a table entry from string or object with html representation."""
    if hasattr(obj, '_repr_html_'):
        return f'<td style="text-align:left;">{obj._repr_html_()}</td>'
    elif hasattr(obj, '_repr_image_svg_xml'):
        return f'<td style="text-align:left;">{obj._repr_image_svg_xml()}</td>'
    else:
        return f'<td style="text-align:left;">{obj}</td>' # uses __str__


# Cell
def tr(cells):
    """Create a table row"""
    tds = [td(cell) for cell in cells]
    return "<tr>\n" + "\n".join(tds) + "\n</tr>"


def th(cells):
    """Create a table header"""
    return tr(cells).replace("<td ", "<th ").replace("</td>", "</th>")


# Cell
def table(rows, header="", cls: str = "gtbook_table"):
    """Create a table from rows"""
    return "<table width=\"100%\" class={cls}>\n" + header + "\n" + "\n".join(rows) + "</table>\n"

def row(cells):
    """Create a table with a single row"""
    return table([tr(cells)])

def ROW(cells): return HTML(row(cells))


# AUTOGENERATED! DO NOT EDIT! File to edit: drone.ipynb (unless otherwise specified).

__all__ = ['axes_figure', 'axes']

# Cell
import math
import numpy as np
import plotly.express as px
import pandas as pd
import gtsam

# Cell
def axes_figure(pose: gtsam.Pose3, scale: float = 1.0, labels: list = ["X", "Y", "Z"]):
    """Create plotly express figure with Pose3 coordinate frame."""
    t = np.reshape(pose.translation(),(3,1))
    M = np.hstack([t,t,t,pose.rotation().matrix() * scale + t])
    df = pd.DataFrame({"x": M[0], "y": M[1], "z": M[2]}, labels+labels)
    return px.line_3d(df, x="x", y="y", z="z", color=df.index,
                     color_discrete_sequence=["red", "green", "blue"])

def axes(*args, **kwargs):
    """Create 3 Scatter3d traces representing Pose3 coordinate frame."""
    return axes_figure(*args, **kwargs).data


# AUTOGENERATED! DO NOT EDIT! File to edit: display.ipynb (unless otherwise specified).

__all__ = ['show', 'pretty', 'randomImages', 'ROBOTS']

# Cell
import math
import gtsam
import graphviz
from .discrete import Variables
from .dbn import dbn_writer, has_positions


# Cell
class show(graphviz.Source):
    """ Display an object with a dot method as a graph."""

    def __init__(self, obj, *args, **kwargs):
        """Construct from object with 'dot' method."""
        # This small class takes an object, calls its dot function, and uses the
        # resulting string to initialize a graphviz.Source instance. This in turn
        # has a _repr_mimebundle_ method, which then renders it in the notebook.
        engine = "dot"

        # Check for Variables argument
        if args and isinstance(args[0], Variables):
            assert len(args) == 1, "Variables must be only positional argument."
            args = [args[0].keyFormatter()]

        # Check if a DotWriter needs creating (or amending)
        writer = dbn_writer(**kwargs)
        if writer is not None:
            kwargs = {"writer": writer}
            if has_positions(writer):
                engine = "neato"
        else:
            kwargs = {}

        # also switch to neato if NonlinearFactprGraph, as needs values always
        if isinstance(obj, gtsam.NonlinearFactorGraph):
            engine = "neato"
        super().__init__(obj.dot(*args, **kwargs), engine=engine)


# Cell
class pretty:
    """Render an object as html with optional arguments."""

    def __init__(self, obj, *args):
        if args and isinstance(args[0], Variables):
            assert len(args) == 1, "Variables must be only argument."
            variables = args[0]
            if isinstance(obj, gtsam.DiscreteValues):
                self._html = variables.values_html(obj)
            else:
                self._html = obj._repr_html_(
                    variables.keyFormatter(), variables.names())
        else:
            if isinstance(obj, gtsam.DiscreteValues):
                self._html = f"{obj}"
            elif isinstance(obj, gtsam.Pose2):
                x = round(obj.x(), 2)
                y = round(obj.y(), 2)
                theta = round(math.degrees(obj.theta()), 2)
                self._html = f"(x={x}, y={y}, theta={theta})"
            else:
                self._html = obj._repr_html_(*args)

    def _repr_html_(self):
        return self._html


# Cell
from IPython.display import HTML
import random

ROBOTS = ["Robot%20menagerie", "Trash%20sorting%20robot%20with%20gripper", "iRobot%20vacuuming%20robot", "Warehouse%20robots",
          "Two-wheeled%20Toy%20Robot", "Autonomous%20Vehicle%20with%20LIDAR%20and%20cameras", "Autonomous%20camera%20drone"]


def randomImages(ch: int, sec: int, style: str, nrImages: int, maxIndex: int = 8):
    """Create an HTML element with some random images.

    Args:
        ch (int): chapter number (base 1)
        sec (int): section number
        style (str): "cubist" | "steampunk" | "expressive"
        nrImages (int): number of images
        maxIndex (int, optional): Indexes to sample from. Defaults to 8.

    Returns:
        HTML: div with nrImages
    """
    perc = round(100/nrImages)
    robot = ROBOTS[ch-1]

    def image_tag(index):
        url = f"https://github.com/gtbook/robotics/blob/main/Art/{style}/S{ch}{sec}-{robot}-{index:02d}.jpg?raw=1"
        return f"<img src='{url}' style='height:256 width:{perc}%'/>"
    indices = random.sample(range(maxIndex), nrImages)
    return HTML(f"""
        <div align='center'>
        {" ".join([image_tag(index) for index in indices])}
        </div>
        """)



# AUTOGENERATED! DO NOT EDIT! File to edit: diffdrive.ipynb (unless otherwise specified).

__all__ = ['axes', 'plane', 'ray', 'show_3d', 'read_image', 'conv2']

# Cell
import gtsam
import math
import PIL
import numpy as np
import plotly.express as px
import plotly.graph_objects as go

# Cell
def axes(length=1.5):
    """Add axes to a plotly figure"""
    return [
        go.Scatter3d(x=[0,length], y=[0,0], z=[0,0], mode="lines", line=dict(color="red")),
        go.Scatter3d(x=[0,0], y=[0,length], z=[0,0], mode="lines", line=dict(color="green")),
        go.Scatter3d(x=[0,0], y=[0,0], z=[0,length], mode="lines", line=dict(color="blue")),
        go.Scatter3d(x=[0], y=[0], z=[0], mode="markers", marker=dict(size=3, color="cyan")),
    ]

def plane(Z:float):
    return [
            go.Scatter3d(x=[-1.2, 1.2], y=[-1,-1], z=[Z, Z], mode="lines", line=dict(color="blue")),
            go.Scatter3d(x=[-1.2, 1.2], y=[ 1, 1], z=[Z, Z], mode="lines", line=dict(color="blue")),
            go.Scatter3d(x=[-1.2,-1.2], y=[-1, 1], z=[Z, Z], mode="lines", line=dict(color="blue")),
            go.Scatter3d(x=[ 1.2, 1.2], y=[-1, 1], z=[Z, Z], mode="lines", line=dict(color="blue")),
    ]

def ray(point3, F, color="orange"):
    X,Y,Z = point3
    x,y = F*X/Z, F*Y/Z
    return go.Scatter3d(x=[X,0,x], y=[Y,0,y], z=[Z,0,F], marker=dict(size=3, color=color))

def show_3d(fig):
    camera = dict(up=dict(x=0, y=-1, z=0), center=dict(x=0, y=0, z=0), eye=dict(x=1.7, y=-0.3, z=-1.1))
    fig.update_layout(scene_camera=camera, margin=dict(t=0, r=0, l=0, b=0), title="Pinhole Camera Model", showlegend=False)
    fig.show()


# Cell
import PIL
import requests

def read_image(image_name):
    """Read image from a the book repo"""
    url = f"https://raw.githubusercontent.com/gtbook/robotics/main/Figures5/{image_name}"
    return PIL.Image.open(requests.get(url, stream=True).raw)

# Cell
import torch

def conv2(input, filter):
    """Convolve input image of shape (iW,iH) with filter of shape (kW,kH)"""
    iW, iH = input.shape
    kW, kH = filter.shape
    _input = torch.reshape(input, (1, 1, iW, iH))
    _filter = torch.reshape(filter, (1, 1, kW, kH))
    _output = torch.conv2d(_input, _filter, padding='same')
    return torch.reshape(_output, (iW, iH))


# AUTOGENERATED! DO NOT EDIT! File to edit: dbn.ipynb (unless otherwise specified).

__all__ = ['dbn_writer', 'has_positions', 'Axis']

# Cell
import gtsam

# Cell
from gtsam import GraphvizFormatting
Axis = GraphvizFormatting.Axis

def dbn_writer(writer=None, hints: dict = None, positions: dict = None,
               boxes: set = None, factor_positions: dict = None,
               binary_edges=False, **kwargs):
    """ Create a DotWriter depending on input arguments:
        If writer is supplied, we will add but not overwrite hints or positions.
    """
    if writer is None and hints is None and positions is None and boxes is None and factor_positions is None and binary_edges==False:
        return None
    writer = GraphvizFormatting() if writer is None else writer

    writer.paperHorizontalAxis = Axis.X
    writer.paperVerticalAxis = Axis.Y

    # Copy hints without overwriting
    if hints is not None:
        assert isinstance(hints, dict)
        ph: dict = writer.positionHints
        for key, y in hints.items():
            if key not in ph:
                ph[key] = y
        writer.positionHints = ph
    # Copy positions without overwriting
    if positions is not None:
        assert isinstance(positions, dict)
        kp: dict = writer.variablePositions
        for key, position in positions.items():
            if key not in kp:
                kp[key] = position
        writer.variablePositions = kp
    # Add boxes
    if boxes is not None:
        assert isinstance(boxes, set)
        bx: set = writer.boxes
        for key in boxes:
            bx.add(key)
        writer.boxes = bx
    # Copy factor positions without overwriting
    if factor_positions is not None:
        assert isinstance(factor_positions, dict)
        kp: dict = writer.factorPositions
        for i, position in factor_positions.items():
            if i not in kp:
                kp[i] = position
        writer.factorPositions = kp
    writer.binaryEdges = binary_edges
    return writer


def has_positions(writer):
    """Check if writer has positions for engine selection"""
    if writer is None:
        return False
    return len(writer.positionHints) > 0 or len(writer.variablePositions) > 0 or len(writer.factorPositions) > 0



# AUTOGENERATED! DO NOT EDIT! File to edit: discrete.ipynb (unless otherwise specified).

__all__ = ['Variables', 'DiscreteKey']

# Cell
import gtsam

from typing import List, Tuple, Callable, Dict, Iterable


# Cell

DiscreteKey = Tuple[int, int]


class Variables:
    """A problem domain for discrete variables."""

    BINARY = ["false", "true"]

    def __init__(self):
        """Default constructor"""
        self._variables = dict()

    def size(self):
        """Return number of variables defined."""
        return len(self._variables)

    def discrete(self, name: str, domain: List[str]) -> DiscreteKey:
        """Create a variable with given name and discrete domain of named values.

        Args:
            name (str): name of the variable.
            domain (List[str]): names for the different values.

        Returns:
            DiscreteKey, i.e., (gtsam.Key, cardinality)
        """
        discreteKey = len(self._variables), len(domain)
        self._variables[discreteKey[0]] = name, domain
        return discreteKey

    def discrete_series(self, character: str, indices: Iterable[int],
                        domain: List[str]) -> Dict[int, DiscreteKey]:
        """Create several discrete variables with Symbol names.

        Args:
            character (str): a single character.
            indices: (Iterable[int]): a set of integer indices.
            domain (List[str]): names for the different values.

        Returns:
            Dict[int, DiscreteKey], i.e., [(gtsam.Key, cardinality)]
        """
        assert len(character) == 1, "discrete_series: requires character only"
        n = len(domain)

        def make_key(index):
            symbol = gtsam.Symbol(character, index)
            key = symbol.key()
            name = symbol.string()
            self._variables[key] = name, domain
            return (key, n)

        return {index: make_key(index) for index in indices}

    def binary(self, name: str) -> DiscreteKey:
        """Create a binary variable with given name.

        Args:
            name (str): name of the variable.

        Returns:
            DiscreteKey, i.e., (gtsam.Key, cardinality)
        """
        return self.discrete(name, self.BINARY)

    def name(self, discreteKey: DiscreteKey) -> str:
        """Return name of variable with given discreteKey.

        Args:
            discreteKey (DiscreteKey): (gtsam.Key, cardinality)

        Returns:
            str: name of the variable.
        """
        return self._variables[discreteKey[0]][0]

    def domain(self, discreteKey: DiscreteKey):
        """Return domain of variable with given discreteKey.

        Args:
            discreteKey (DiscreteKey): (gtsam.Key, cardinality)

        Returns:
            str: domain of the variable.
        """
        return self._variables[discreteKey[0]][1]

    def keyFormatter(self) -> Callable:
        """Return a lambda that can be used as KeyFormatter in GTSAM"""
        return lambda key: self._variables[key][0]

    def names(self) -> Dict[int, List[str]]:
        """Return a names dictionary that is used by GTSAM markdown methods"""
        return {key: domain for (key, (name, domain)) in self._variables.items()}

    def assignment(self, map: Dict[DiscreteKey, str]) -> gtsam.DiscreteValues:
        """Create a GTSAM assignment of keys to values.

        Args:
            map (Dict[DiscreteKey, str]): map from discrete keys to values.

        Returns:
            gtsam.DiscreteValues: the GTSAM equivalent.
        """
        values = gtsam.DiscreteValues()
        for discreteKey, value in map.items():
            domain = self.domain(discreteKey)
            assert value in domain, f"Specified value '{value}' not found in domain of '{self.name(discreteKey)}'. Allowed values are {domain}."
            values[discreteKey[0]] = domain.index(value)
        return values

    def values_markdown(self, assignment: gtsam.DiscreteValues) -> str:
        """Render a DiscreteValues instance as markdown.

        Args:
            assignment (gtsam.DiscreteValues): the values to render.

        Returns:
            str: a markdown string.
        """
        return gtsam.markdown(assignment, self.keyFormatter(), self.names())

    def values_html(self, assignment: gtsam.DiscreteValues) -> str:
        """Render a DiscreteValues instance as html.

        Args:
            assignment (gtsam.DiscreteValues): the values to render.

        Returns:
            str: a html string.
        """
        return gtsam.html(assignment, self.keyFormatter(), self.names())

    def _repr_html_(self):
        """An html representation
        """
        # Print out preamble.
        ss = "<div>\n<table class='Variables'>\n  <thead>\n"

        # Print out header row.
        ss += "    <tr><th>Variable</th><th>Domain</th></tr>\n"

        # Finish header and start body.
        ss += "  </thead>\n  <tbody>\n"

        # Print out all rows.
        for name, domain in self._variables.values():
            ss += "    <tr>"
            ss += "<th>" + name + "</th><td>"
            for value in domain[:-1]:
                ss += value + ", "
            ss += domain[-1]
            ss += "</td>"
            ss += "</tr>\n"
        ss += "  </tbody>\n</table>\n</div>"
        return ss
